import sys
import json
import json.decoder
import urllib.request
import urllib.error
import html.parser
from .stdout import Stdout


class Asciicast:

    def __init__(self, stdout, width, height, duration, command=None, title=None, term=None, shell=None):
        self.stdout = stdout
        self.width = width
        self.height = height
        self.duration = duration
        self.command = command
        self.title = title
        self.term = term
        self.shell = shell

    def save(self, path):
        stdout = list(map(lambda frame: [round(frame[0], 6), frame[1]], self.stdout.frames))
        duration = round(self.duration, 6)
        attrs = {
            "version": 1,
            "width": self.width,
            "height": self.height,
            "duration": duration,
            "command": self.command,
            "title": self.title,
            "env": {
                "TERM": self.term,
                "SHELL": self.shell
            },
            "stdout": stdout
        }

        with open(path, "w") as f:
            f.write(json.dumps(attrs, ensure_ascii=False, indent=2))


# asciinema play file.json
# asciinema play https://asciinema.org/a/123.json
# asciinema play https://asciinema.org/a/123
# asciinema play ipfs://ipfs/QmbdpNCwqeZgnmAWBCQcs8u6Ts6P2ku97tfKAycE1XY88p
# asciinema play -


class LoadError(Exception):
    pass


class Parser(html.parser.HTMLParser):
    def __init__(self):
        html.parser.HTMLParser.__init__(self)
        self.url = None

    def handle_starttag(self, tag, attrs_list):
        # look for <link rel="alternate" type="application/asciicast+json" href="https://...json">
        if tag == 'link':
            attrs = {}
            for k, v in attrs_list:
                attrs[k] = v

            if attrs.get('rel') == 'alternate' and attrs.get('type') == 'application/asciicast+json':
                self.url = attrs.get('href')


def fetch(url):
    if url.startswith("ipfs:/"):
        url = "https://ipfs.io/%s" % url[6:]
    elif url.startswith("fs:/"):
        url = "https://ipfs.io/%s" % url[4:]

    if url == "-":
        return sys.stdin.read()

    if url.startswith("http:") or url.startswith("https:"):
        response = urllib.request.urlopen(url)
        data = response.read().decode(errors='replace')

        content_type = response.headers['Content-Type']
        if content_type and content_type.startswith('text/html'):
            parser = Parser()
            parser.feed(data)
            url = parser.url

            if not url:
                raise LoadError("""<link rel="alternate" type="application/asciicast+json" href="..."> not found in fetched HTML document""")

            return fetch(url)

        return data

    with open(url, 'r') as f:
        return f.read()


def load(filename):
    try:
        attrs = json.loads(fetch(filename))

        if type(attrs) != dict:
            raise LoadError('unsupported asciicast format')

        return Asciicast(
            attrs['stdout'],
            attrs['width'],
            attrs['height'],
            attrs['duration'],
            attrs['command'],
            attrs['title']
        )
    except (OSError, urllib.error.HTTPError) as e:
        raise LoadError(str(e))
    except json.decoder.JSONDecodeError as e:
        raise LoadError('JSON decoding error: ' + str(e))
    except KeyError as e:
        raise LoadError('asciicast is missing key ' + str(e))



from asciinema.commands.command import Command
from asciinema.player import Player
import asciinema.asciicast as asciicast


class PlayCommand(Command):

    def __init__(self, filename, max_wait, speed, player=None):
        Command.__init__(self)
        self.filename = filename
        self.max_wait = max_wait
        self.speed = speed
        self.player = player if player is not None else Player()

    def execute(self):
        try:
            self.player.play(asciicast.load(self.filename), self.max_wait, self.speed)

        except asciicast.LoadError as e:
            self.print_warning("Playback failed: %s" % str(e))
            return 1
        except KeyboardInterrupt:
            return 1

        return 0



"""
This code from asciinema project
asciinema homepage: https://asciinema.org
asciinema github: https://github.com/asciinema/asciinema.git

Trieu Pham (trieu@codelyn.io) have modified it for purpose of project xirion
"""


import locale
import argparse
import os
from os.path import expanduser
import sys

import inspect

# from asciinema import __version__
from codelynx import __version__
import asciinema.config as config
from asciinema.commands.auth import AuthCommand
from asciinema.commands.record import RecordCommand
from asciinema.commands.play import PlayCommand
from asciinema.commands.upload import UploadCommand
from asciinema.api import Api

from codelynx.commands.lookup import Lookup
from codelynx.commands.ask import Ask

home = expanduser("~")

def positive_float(value):

    value = float(value)
    if value <= 0.0:
        raise argparse.ArgumentTypeError("must be positive")

    return value

def config_server(args):
    # try:
    fo = open("conf", "w+")
    fo.write(args.url)
    fo.close()
    # except Exception as ex:
    #     print(ex)
    #     return 1
    return 0

def insert_lookup(args, config):
    return Lookup()


def xirion_ask(args):
    return Ask(args.question)


def rec_command(args, config):

    api = Api(config.api_url, os.environ.get("USER"), config.api_token)

    return RecordCommand(api, args.filename, args.command, args.title, args.yes, args.quiet, args.max_wait)


def play_command(args, config):
    return PlayCommand(args.filename, args.max_wait, args.speed)


def upload_command(args, config):
    api = Api(config.api_url, os.environ.get("USER"), config.api_token)
    return UploadCommand(api, args.filename)


def auth_command(args, config):
    return AuthCommand(config.api_url, config.api_token)


def maybe_str(v):
    if v is not None:
        return str(v)


def main():

    if locale.nl_langinfo(locale.CODESET).upper() != 'UTF-8':
        # print("asciinema needs a UTF-8 native locale to run. Check the output of `locale` command.")
        sys.exit(1)

    cfg = config.load()

    # create the top-level parser
    parser = argparse.ArgumentParser(
        description="Record and share your terminal sessions, the right way.",
        epilog="""
            example usage:
            Record terminal and upload it to asciinema.org:
                \x1b[1masciinema rec\x1b[0m
            Record terminal to local file:
                \x1b[1masciinema rec demo.json\x1b[0m
            Record terminal and upload it to asciinema.org, specifying title:
                \x1b[1masciinema rec -t "My git tutorial"\x1b[0m
            Record terminal to local file, "trimming" longer pauses to max 2.5 sec:
                \x1b[1masciinema rec -w 2.5 demo.json\x1b[0m
            Replay terminal recording from local file:
                \x1b[1masciinema play demo.json\x1b[0m
            Replay terminal recording hosted on asciinema.org:
                \x1b[1masciinema play https://asciinema.org/a/difqlgx86ym6emrmd8u62yqu8\x1b[0m

            For help on a specific command run:
            \x1b[1masciinema <command> -h\x1b[0m""",
        formatter_class=argparse.RawDescriptionHelpFormatter
    )
    parser.add_argument('--version',
                        action='version', version='xirion %s' % __version__)

    subparsers = parser.add_subparsers()

    # create the parser for "lookup" command
    parser_lookup = subparsers.add_parser(
        'lookup', help='Build lookup collection')
    parser_lookup.set_defaults(func=insert_lookup)

    # create parser for config server url
    parser_conf = subparsers.add_parser('config', help="Config server")
    parser_conf.add_argument('url', nargs='?', help="server url")
    parser_conf.set_defaults(func=config_server)

    # create the parser for "ask" command
    parser_ask = subparsers.add_parser('ask', help="Ask xirion")
    parser_ask.add_argument('question', nargs='?',
                            help="ask xirion this question")
    parser_ask.set_defaults(func=xirion_ask)

    # create the parser for the "rec" command
    parser_rec = subparsers.add_parser('rec', help='Record terminal session')
    parser_rec.add_argument('-c', '--command',
                            help='command to record, defaults to $SHELL', default=cfg.record_command)
    parser_rec.add_argument('-t', '--title',
                            help='title of the asciicast')
    parser_rec.add_argument('-w', '--max-wait',
                            help='limit recorded terminal inactivity to max <sec> seconds (can be fractional)',
                            type=positive_float, default=maybe_str(cfg.record_max_wait))
    parser_rec.add_argument('-y', '--yes',
                            help='answer "yes" to all prompts (e.g. upload confirmation)',
                            action='store_true', default=cfg.record_yes)
    parser_rec.add_argument('-q', '--quiet',
                            help='be quiet, suppress all notices/warnings (implies -y)',
                            action='store_true', default=cfg.record_quiet)
    parser_rec.add_argument('filename',
                            nargs='?', default='',
                            help='filename/path to save the recording to')
    parser_rec.set_defaults(func=rec_command)

    # create the parser for the "play" command
    parser_play = subparsers.add_parser('play', help='Replay terminal session')
    parser_play.add_argument('-w', '--max-wait',
                             help='limit terminal inactivity to max <sec> seconds (can be fractional)',
                             type=positive_float, default=maybe_str(cfg.play_max_wait))
    parser_play.add_argument('-s', '--speed',
                             help='playback speedup (can be fractional)',
                             type=positive_float, default=cfg.play_speed)
    parser_play.add_argument('filename',
                             help='local path, http/ipfs URL or "-" (read from stdin)')
    parser_play.set_defaults(func=play_command)

    # create the parser for the "upload" command
    parser_upload = subparsers.add_parser('upload',
                                          help='Upload locally saved terminal session to asciinema.org')
    parser_upload.add_argument('filename',
                               help='filename or path of local recording')
    parser_upload.set_defaults(func=upload_command)

    # create the parser for the "auth" command
    parser_auth = subparsers.add_parser('auth',
                                        help='Manage recordings on asciinema.org account')
    parser_auth.set_defaults(func=auth_command)

    # parse the args and call whatever function was selected
    args = parser.parse_args()

    if hasattr(args, 'question'):
        command = args.func(args)
        code = command.execute()
        sys.exit(code)

    if hasattr(args, 'func'):
        if hasattr(args, 'question'):
            command = args.func(args)

        elif hasattr(args, 'url'):
            code = args.func(args)
            sys.exit(code)
        else:
            command = args.func(args, cfg)
        code = command.execute()
        sys.exit(code)
    else:
        parser.print_help()
        sys.exit(1)


if __name__ == '__main__':
    main()



import sys
import os
import tempfile

from asciinema.commands.command import Command
from asciinema.recorder import Recorder
from asciinema.api import APIError


class RecordCommand(Command):

    def __init__(self, api, filename, command, title, assume_yes, quiet, max_wait, recorder=None):
        Command.__init__(self, quiet)
        self.api = api
        self.filename = filename
        self.command = command
        self.title = title
        self.assume_yes = assume_yes or quiet
        self.max_wait = max_wait
        self.recorder = recorder if recorder is not None else Recorder()

    def execute(self):
        if self.filename == "":
            self.filename = _tmp_path()
            upload = True
        else:
            upload = False

        try:
            _touch(self.filename)
        except OSError as e:
            self.print_warning("Can't record to %s: %s" % (self.filename, str(e)))
            return 1

        self.print_info("Asciicast recording started.")
        self.print_info("""Hit Ctrl-D or type "exit" to finish.""")

        self.recorder.record(self.filename, self.command, self.title, self.max_wait)

        self.print_info("Asciicast recording finished.")

        if upload:
            if not self.assume_yes:
                self.print_info("Press <Enter> to upload, <Ctrl-C> to cancel.")
                try:
                    sys.stdin.readline()
                except KeyboardInterrupt:
                    return 0

            try:
                url, warn = self.api.upload_asciicast(self.filename)
                if warn:
                    self.print_warning(warn)
                os.remove(self.filename)
                self.print(url)
            except APIError as e:
                self.print_warning("Upload failed: %s" % str(e))
                self.print_warning("Retry later by running: asciinema upload %s" % self.filename)
                return 1

        return 0


def _tmp_path():
    fd, path = tempfile.mkstemp(suffix='-asciinema.json')
    os.close(fd)
    return path


def _touch(path):
    open(path, 'a').close()



"""
Author: Trieu Pham (trieu@codelynx.io)

A private property of #codelynx company
"""

from codelynx.api import Api
from asciinema.commands.command import Command


class Lookup(Command):
    """
    This class manages insert new document to lookup collection
    Expected command: xirion lookup
    Response: switch to insert mode
    @args: None
    @return:
        0 if everything 's good, 1 otherwise
    @raises:
        unknown
    """
    def __init__(self):
        # self.cmd = cmd
        # self.correction = correction
        self.api = Api()
        Command.__init__(self)  # color

    def execute(self):
        try:
            cmd = None
            while True:
                cmd = input('Command: ')
                if cmd == 'exit':
                    self.print_info('Bye! exit insert lookup mode ~')
                    break
                correction = input('Correction: ')
                # Save these data to server
                # then make some noise ~ print response
                # kwargs = dict(orig=cmd, correct=correction)
                self.api.handle_data('lookup', quiet=False,
                                   orig=cmd, correct=correction)
        except Exception as ex:
            print(ex)  # For debug
            return 1  # exit code 1 ~ error
        return 0  # if code runs to this line, it means everything is ok => exit code 0



import codecs
import mimetypes
import sys
import uuid
import io
import base64

from urllib.request import Request, urlopen
from urllib.error import HTTPError, URLError
from .http_adapter import HTTPConnectionError


class MultipartFormdataEncoder:
    def __init__(self):
        self.boundary = uuid.uuid4().hex
        self.content_type = 'multipart/form-data; boundary={}'.format(self.boundary)

    @classmethod
    def u(cls, s):
        if sys.hexversion >= 0x03000000 and isinstance(s, bytes):
            s = s.decode('utf-8')
        return s

    def iter(self, fields, files):
        """
        fields is a dict of {name: value} for regular form fields.
        files is a dict of {name: (filename, file-type)} for data to be uploaded as files
        Yield body's chunk as bytes
        """
        encoder = codecs.getencoder('utf-8')
        for (key, value) in fields.items():
            key = self.u(key)
            yield encoder('--{}\r\n'.format(self.boundary))
            yield encoder(self.u('Content-Disposition: form-data; name="{}"\r\n').format(key))
            yield encoder('\r\n')
            if isinstance(value, int) or isinstance(value, float):
                value = str(value)
            yield encoder(self.u(value))
            yield encoder('\r\n')
        for (key, filename_and_f) in files.items():
            filename, f = filename_and_f
            key = self.u(key)
            filename = self.u(filename)
            yield encoder('--{}\r\n'.format(self.boundary))
            yield encoder(self.u('Content-Disposition: form-data; name="{}"; filename="{}"\r\n').format(key, filename))
            yield encoder('Content-Type: {}\r\n'.format(mimetypes.guess_type(filename)[0] or 'application/octet-stream'))
            yield encoder('\r\n')
            data = f.read()
            yield (data, len(data))
            yield encoder('\r\n')
        yield encoder('--{}--\r\n'.format(self.boundary))

    def encode(self, fields, files):
        body = io.BytesIO()
        for chunk, chunk_len in self.iter(fields, files):
            body.write(chunk)
        return self.content_type, body.getvalue()


class URLLibHttpAdapter:

    def post(self, url, fields={}, files={}, headers={}, username=None, password=None):
        content_type, body = MultipartFormdataEncoder().encode(fields, files)

        headers = headers.copy()
        headers["Content-Type"] = content_type

        if password:
            auth = "%s:%s" % (username, password)
            encoded_auth = base64.encodestring(auth.encode('utf-8'))[:-1]
            headers["Authorization"] = b"Basic %" + encoded_auth

        request = Request(url, data=body, headers=headers, method="POST")

        try:
            response = urlopen(request)
            status = response.status
            headers = self._parse_headers(response)
            body = response.read().decode('utf-8')
        except HTTPError as e:
            status = e.code
            headers = {}
            body = e.read().decode('utf-8')
        except URLError as e:
            raise HTTPConnectionError(str(e))

        return (status, headers, body)

    def _parse_headers(self, response):
        headers = {}
        for k, v in response.getheaders():
            headers[k] = v

        return headers



import errno
import os
import pty
import signal
import tty
import array
import fcntl
import termios
import select
import io
import shlex
import sys
import struct
import re
import requests
import json

from asciinema.term import raw
from codelynx.api import Api


class PtyRecorder:

    def record_command(self, command, output, env=os.environ):

        master_fd = None
        # Read buffer, max = 1024
        rbuf = 1024
        get_ecode = False
        cmdlast = ''
        my_data = ''
        ecode = ''

        def _set_pty_size():
            '''
            Sets the window size of the child pty based on the window size
            of our own controlling terminal.
            '''

            # Get the terminal size of the real terminal, set it on the
            # pseudoterminal.
            if os.isatty(pty.STDOUT_FILENO):
                buf = array.array('h', [0, 0, 0, 0])
                fcntl.ioctl(pty.STDOUT_FILENO, termios.TIOCGWINSZ, buf, True)
                fcntl.ioctl(master_fd, termios.TIOCSWINSZ, buf)
            else:
                buf = array.array('h', [24, 80, 0, 0])
                fcntl.ioctl(master_fd, termios.TIOCSWINSZ, buf)

        def _write_stdout(data):
            '''
            Writes to stdout as if the child process had written the data.
            '''

            # collect environment info
            session_id = os.environ._data[b'TERM_SESSION_ID'].decode()
            # google_api_key = os.environ._data[b'GOOGLE_API_KEY'].decode()
            terminal_user = os.environ._data[b'USER'].decode()
            shell = os.environ._data[b'SHELL'].decode()
            term_program = os.environ._data[b'TERM_PROGRAM'].decode()

            # post data to server
            codelynx_api.handle_data('data', quiet=True, rec=data.decode(),
                                     session_id=session_id, shell=shell,
                                     terminal_user=terminal_user,
                                     term_program=term_program)
            # print data on terminal
            os.write(pty.STDOUT_FILENO, data)
            # if cmdlast == b'\r':
            #     nonlocal my_data
            #     my_data += str(data, 'utf-8')

            #     if my_data.endswith('$ '):
            #         my_data = re.split(r'\r\n', my_data)
            #         nonlocal ecode
            #         data_length = len(my_data)
            #         ecode = my_data.pop(data_length - 2)
            #         # Remove echo line
            #         del my_data[data_length - 3]
            #         my_data = "\r\n".join(my_data)

            #         # Send data to server
            #         codelynx_api.handle_data('', my_data)
            #         # Print command output
            #         os.write(pty.STDOUT_FILENO, my_data.encode())
            # else:
            #     os.write(pty.STDOUT_FILENO, data)

        def _handle_master_read(data):
            nonlocal my_data
            '''Handles new data on child process stdout.'''

            # Add time, decoder to data
            _write_stdout(data)

        def _write_master(data):
            '''Writes to the child process from its controlling terminal.'''

            # Record input char
            # codelynx_api.handle_data(str(data, 'utf-8'), '')

            while data:
                n = os.write(master_fd, data)
                data = data[n:]

        def _handle_stdin_read(data):
            '''Handles new data on child process stdin.'''
            _write_master(data)

        def _signals(signal_list):
            old_handlers = []
            for sig, handler in signal_list:
                old_handlers.append((sig, signal.signal(sig, handler)))
            return old_handlers

        def _copy(signal_fd):
            '''Main select loop.

            Passes control to _master_read() or _stdin_read()
            when new data arrives.
            '''

            fds = [master_fd, pty.STDIN_FILENO, signal_fd]
            cmdin = ''
            cmdout = ''
            cmdcode = ''
            nonlocal cmdlast
            nonlocal my_data

            while True:
                try:
                    rfds, wfds, xfds = select.select(fds, [], [])

                except OSError as e:  # Python >= 3.3
                    if e.errno == errno.EINTR:
                        continue
                except select.error as e:  # Python < 3.3
                    if e.args[0] == 4:
                        continue

                # Output
                if master_fd in rfds:
                    if cmdlast == b'\r':
                        data = os.read(master_fd, rbuf)
                        cmdin = ''
                        cmdout = str(data, 'utf-8')

                    else:
                        data = os.read(master_fd, rbuf)

                    if not data:  # Reached EOF.
                        fds.remove(master_fd)
                    else:
                        _handle_master_read(data)

                # Input
                if pty.STDIN_FILENO in rfds:
                    data = os.read(pty.STDIN_FILENO, rbuf)
                    nonlocal my_data
                    my_data = str(data, 'utf-8')

                    cmdlast = data
                    # logic: turn exe signal to exe and echo exit code of last
                    # cmd
                    # if data == b'\r':
                    #     data = b'\recho $?\r'

                    cmdin += str(data, 'utf-8')
                    if not data:
                        fds.remove(pty.STDIN_FILENO)
                    else:
                        _handle_stdin_read(data)

                if signal_fd in rfds:
                    data = os.read(signal_fd, rbuf)
                    if data:
                        signals = struct.unpack('%uB' % len(data), data)
                        for sig in signals:
                            if sig in [signal.SIGCHLD,
                                       signal.SIGHUP,
                                       signal.SIGTERM,
                                       signal.SIGQUIT]:
                                os.close(master_fd)
                                return
                            elif sig == signal.SIGWINCH:
                                _set_pty_size()

        pid, master_fd = pty.fork()

        if pid == pty.CHILD:
            os.execvpe(command[0], command, env)

        pipe_r, pipe_w = os.pipe()
        flags = fcntl.fcntl(pipe_w, fcntl.F_GETFL, 0)
        flags = flags | os.O_NONBLOCK
        flags = fcntl.fcntl(pipe_w, fcntl.F_SETFL, flags)

        signal.set_wakeup_fd(pipe_w)

        old_handlers = _signals(map(lambda s: (s, lambda signal, frame: None),
                                    [signal.SIGWINCH,
                                     signal.SIGCHLD,
                                     signal.SIGHUP,
                                     signal.SIGTERM,
                                     signal.SIGQUIT]))

        _set_pty_size()

        with raw(pty.STDIN_FILENO):
            try:
                codelynx_api = Api()
                _copy(pipe_r)
            except (IOError, OSError):
                pass

        _signals(old_handlers)

        os.waitpid(pid, 0)
        output.close()



import os
import sys
import time

from asciinema.term import raw, read_non_blocking


class Player:

    def play(self, asciicast, max_wait=None, speed=1.0):
        if os.isatty(sys.stdin.fileno()):
            with raw(sys.stdin.fileno()):
                self._play(asciicast, max_wait, speed, True)
        else:
            self._play(asciicast, max_wait, speed, False)

    def _play(self, asciicast, max_wait, speed, raw):
        for delay, text in asciicast.stdout:
            if max_wait and delay > max_wait:
                delay = max_wait
            time.sleep(delay / speed)
            sys.stdout.write(text)
            sys.stdout.flush()

            if raw:
                data = read_non_blocking(sys.stdin.fileno())
                if 0x03 in data:  # ctrl-c
                    break
